@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008 Andrew G. Morgan <morgan@kernel.org>
+ * Copyright (c) 2008-11 Andrew G. Morgan <morgan@kernel.org>
  *
  * This is a simple 'bash' wrapper program that can be used to
  * raise and lower both the bset and pI capabilities before invoking
@@ -19,22 +19,30 @@
 #include <grp.h>
 #include <errno.h>
 #include <ctype.h>
-#include <sys/wait.h>
 #include <sys/capability.h>
-
-/* prctl based API for altering character of current process */
-#define PR_GET_KEEPCAPS    7
-#define PR_SET_KEEPCAPS    8
-#define PR_CAPBSET_READ   23
-#define PR_CAPBSET_DROP   24
-#define PR_GET_SECUREBITS 27
-#define PR_SET_SECUREBITS 28
+#include <sys/securebits.h>
+#include <sys/wait.h>
+#include <sys/prctl.h>
 
 #define MAX_GROUPS       100   /* max number of supplementary groups for user */
 
 static const cap_value_t raise_setpcap[1] = { CAP_SETPCAP };
 static const cap_value_t raise_chroot[1] = { CAP_SYS_CHROOT };
 
+static char *binary(unsigned long value)
+{
+    static char string[8*sizeof(unsigned long) + 1];
+    unsigned i;
+
+    i = sizeof(string);
+    string[--i] = '\0';
+    do {
+	string[--i] = (value & 1) ? '1' : '0';
+	value >>= 1;
+    } while ((i > 0) && value);
+    return string + i;
+}
+
 int main(int argc, char *argv[], char *envp[])
 {
     pid_t child;
@@ -70,31 +78,48 @@ int main(int argc, char *argv[], char *envp[])
 		exit(1);
 	    }
 
-	    for (ptr = argv[i]+7; (ptr = strtok(ptr, ",")); ptr = NULL) {
-		/* find name for token */
-		cap_value_t cap;
-		int status;
+	    if (strcmp("all", argv[i]+7) == 0) {
+		unsigned j = 0;
+		while (CAP_IS_SUPPORTED(j)) {
+		    if (cap_drop_bound(j) != 0) {
+			char *name_ptr;
 
-		if (cap_from_name(ptr, &cap) != 0) {
-		    fprintf(stderr, "capability [%s] is unknown to libcap\n",
-			    ptr);
-		    exit(1);
+			name_ptr = cap_to_name(j);
+			fprintf(stderr,
+				"Unable to drop bounding capability [%s]\n",
+				name_ptr);
+			cap_free(name_ptr);
+			exit(1);
+		    }
+		    j++;
 		}
-		if (cap_set_proc(raised_for_setpcap) != 0) {
-		    perror("unable to raise CAP_SETPCAP for BSET changes");
-		    exit(1);
-		}
-		status = prctl(PR_CAPBSET_DROP, cap);
-		if (cap_set_proc(orig) != 0) {
-		    perror("unable to lower CAP_SETPCAP post BSET change");
-		    exit(1);
-		}
-		if (status) {
-		    fprintf(stderr, "failed to drop [%s=%u]\n", ptr, cap);
-		    exit(1);
+	    } else {
+		for (ptr = argv[i]+7; (ptr = strtok(ptr, ",")); ptr = NULL) {
+		    /* find name for token */
+		    cap_value_t cap;
+		    int status;
+
+		    if (cap_from_name(ptr, &cap) != 0) {
+			fprintf(stderr,
+				"capability [%s] is unknown to libcap\n",
+				ptr);
+			exit(1);
+		    }
+		    if (cap_set_proc(raised_for_setpcap) != 0) {
+			perror("unable to raise CAP_SETPCAP for BSET changes");
+			exit(1);
+		    }
+		    status = prctl(PR_CAPBSET_DROP, cap);
+		    if (cap_set_proc(orig) != 0) {
+			perror("unable to lower CAP_SETPCAP post BSET change");
+			exit(1);
+		    }
+		    if (status) {
+			fprintf(stderr, "failed to drop [%s=%u]\n", ptr, cap);
+			exit(1);
+		    }
 		}
 	    }
-
 	    cap_free(raised_for_setpcap);
 	    cap_free(orig);
 	} else if (!memcmp("--inh=", argv[i], 6)) {
@@ -131,7 +156,11 @@ int main(int argc, char *argv[], char *envp[])
 		perror("Out of memory for inh set");
 		exit(1);
 	    }
-	    sprintf(ptr, "%s %s+i", text, argv[i]+6);
+	    if (argv[i][6] && strcmp("none", argv[i]+6)) {
+		sprintf(ptr, "%s %s+i", text, argv[i]+6);
+	    } else {
+		strcpy(ptr, text);
+	    }
 
 	    all = cap_from_text(ptr);
 	    if (all == NULL) {
@@ -420,11 +449,12 @@ int main(int argc, char *argv[], char *envp[])
 
 	    for (cap=0; (cap < 64) && (value >> cap); ++cap) {
 		if (value & (1ULL << cap)) {
-		    const char *ptr;
+		    char *ptr;
 
 		    ptr = cap_to_name(cap);
 		    if (ptr != NULL) {
 			printf("%s%s", sep, ptr);
+			cap_free(ptr);
 		    } else {
 			printf("%s%u", sep, cap);
 		    }
@@ -432,13 +462,29 @@ int main(int argc, char *argv[], char *envp[])
 		}
 	    }
 	    printf("\n");
+        } else if (!memcmp("--supports=", argv[i], 11)) {
+	    cap_value_t cap;
+
+	    if (cap_from_name(argv[i] + 11, &cap) < 0) {
+		fprintf(stderr, "cap[%s] not recognized by library\n",
+			argv[i] + 11);
+		exit(1);
+	    }
+	    if (!CAP_IS_SUPPORTED(cap)) {
+		fprintf(stderr, "cap[%s=%d] not supported by kernel\n",
+			argv[i] + 11, cap);
+		exit(1);
+	    }
 	} else if (!strcmp("--print", argv[i])) {
 	    unsigned cap;
 	    int set, status, j;
 	    cap_t all;
 	    char *text;
 	    const char *sep;
-	    gid_t groups[MAX_GROUPS];
+	    struct group *g;
+	    gid_t groups[MAX_GROUPS], gid;
+	    uid_t uid;
+	    struct passwd *u;
 
 	    all = cap_get_proc();
 	    text = cap_to_text(all, NULL);
@@ -448,24 +494,27 @@ int main(int argc, char *argv[], char *envp[])
 
 	    printf("Bounding set =");
  	    sep = "";
-	    for (cap=0; (set = prctl(PR_CAPBSET_READ, cap)) >= 0; cap++) {
-		const char *ptr;
+	    for (cap=0; (set = cap_get_bound(cap)) >= 0; cap++) {
+		char *ptr;
 		if (!set) {
 		    continue;
 		}
 
 		ptr = cap_to_name(cap);
-		if (ptr == 0) {
+		if (ptr == NULL) {
 		    printf("%s%u", sep, cap);
 		} else {
 		    printf("%s%s", sep, ptr);
+		    cap_free(ptr);
 		}
 		sep = ",";
 	    }
 	    printf("\n");
 	    set = prctl(PR_GET_SECUREBITS);
 	    if (set >= 0) {
-		printf("Securebits: 0%o/0x%x\n", set, set);
+		const char *b;
+		b = binary(set);  /* use verilog convention for binary string */
+		printf("Securebits: 0%o/0x%x/%u'b%s\n", set, set, strlen(b), b);
 		printf(" secure-noroot: %s (%s)\n",
 		       (set & 1) ? "yes":"no",
 		       (set & 2) ? "locked":"unlocked");
@@ -485,14 +534,19 @@ int main(int argc, char *argv[], char *envp[])
 		    printf("[Keepcaps ABI not supported]\n");
 		}
 	    }
-	    printf("uid=%u\n", getuid());
-	    printf("gid=%u\n", getgid());
+	    uid = getuid();
+	    u = getpwuid(uid);
+	    printf("uid=%u(%s)\n", getuid(), u ? u->pw_name : "???");
+	    gid = getgid();
+	    g = getgrgid(gid);
+	    printf("gid=%u(%s)\n", gid, g ? g->gr_name : "???");
 	    printf("groups=");
 	    status = getgroups(MAX_GROUPS, groups);
 	    sep = "";
 	    for (j=0; j < status; j++) {
-	      printf("%s%u", sep, groups[j]);
-	      sep = ",";
+		g = getgrgid(groups[j]);
+		printf("%s%u(%s)", sep, groups[j], g ? g->gr_name : "???");
+		sep = ",";
 	    }
 	    printf("\n");
 	} else if ((!strcmp("--", argv[i])) || (!strcmp("==", argv[i]))) {
@@ -507,6 +561,7 @@ int main(int argc, char *argv[], char *envp[])
 		   "  --help         this message (or try 'man capsh')\n"
 		   "  --print        display capability relevant state\n"
 		   "  --decode=xxx   decode a hex string to a list of caps\n"
+		   "  --supports=xxx exit 1 if capability xxx unsupported\n"
 		   "  --drop=xxx     remove xxx,.. capabilities from bset\n"
 		   "  --caps=xxx     set caps as per cap_from_text()\n"
 		   "  --inh=xxx      set xxx,.. inheritiable set\n"
